############################################################################
#                                                                          #
# Imports & Constants                                                      #
#                                                                          #
############################################################################

import gpudb
import csv

HOST = "127.0.0.1"
PORT = "9191"
CSV = "road_weights.csv"

OPTION_NO_ERROR = {"no_error_if_not_exists": "true"}

TABLE_SRN = "seattle_road_network"

GRAPH_S = TABLE_SRN + "_graph"
GRAPH_S_BSOLVED = GRAPH_S + "_backhaul_routing_solved"

############################################################################
#                                                                          #
# Table Setup Function                                                     #
#                                                                          #
############################################################################

def table_setup():
    """ Setup necessary source tables for graph solver examples. """

    print("")
    print("===========")
    print("TABLE SETUP")
    print("===========")
    print("")

    # Clear any related tables in case they already exist
    kinetica.clear_table(table_name=TABLE_SRN, options=OPTION_NO_ERROR)

    # Define the Seattle road network table type
    srn_columns = [
        ["OriginalEdgeID", "long"],
        ["TwoWay", "int"],
        ["WKTLINE", "string", "wkt"],
        ["time", "float"]
    ]

    # Create the Seattle road network table
    try:
        table_srn_obj = gpudb.GPUdbTable(
            _type=srn_columns,
            name=TABLE_SRN,
            db=kinetica,
            options={}
        )
        print("{} table object successfully created.".format(TABLE_SRN))
    except gpudb.GPUdbException as e:
        print("{} table object creation failure: {}".format(TABLE_SRN, str(e)))

    # Insert records from a local CSV file into the Seattle road network table
    print("Creating records from the " + CSV + " file.")
    srn_data = csv.reader(open(CSV))
    srn_data.next()
    srn_records = []
    for record in srn_data:
        record_data = [long(record[0]), int(record[1]), record[2],
                       float(record[3])]
        srn_records.append(record_data)
    print(
        "Inserting records into {} table. This may take a "
        "while.".format(TABLE_SRN)
    )
    table_srn_obj.insert_records(srn_records)
    print(
        "Number of records inserted into the {} "
        "table: {}".format(TABLE_SRN, table_srn_obj.size())
    )

# end table_setup()

############################################################################
#                                                                          #
# Graph Setup Function                                                     #
#                                                                          #
############################################################################

def graph_setup():
    """ Setup graphs to be used in solver examples. """

    print("")
    print("===========")
    print("GRAPH SETUP")
    print("===========")
    print("")

    # Create a graph from TABLE_SRN
    print("Creating {}".format(GRAPH_S))
    create_s_graph_response = kinetica.create_graph(
        graph_name=GRAPH_S,
        directed_graph=True,
        nodes=[],
        edges=[
            TABLE_SRN + ".WKTLINE AS EDGE_WKTLINE",
            TABLE_SRN + ".TwoWay AS EDGE_DIRECTION"
        ],
        weights=[
            TABLE_SRN + ".WKTLINE AS WEIGHTS_EDGE_WKTLINE",
            TABLE_SRN + ".TwoWay AS WEIGHTS_EDGE_DIRECTION",
            TABLE_SRN + ".time AS WEIGHTS_VALUESPECIFIED"
        ],
        restrictions=[],
        options={
            "recreate": "true"
        }
    )
    print(
        "{} created: {}".format(GRAPH_S,
                                create_s_graph_response["status_info"]
                                ["status"])
    )
    print("Number of nodes: {}".format(create_s_graph_response["num_nodes"]))
    print("Number of edges: {}".format(create_s_graph_response["num_edges"]))

# end graph_setup()

############################################################################
#                                                                          #
# Backhaul Routing Solving Function                                        #
#                                                                          #
############################################################################

def backhaul_example():
    """Demonstrate creating and solving a graph using the BACKHAUL_ROUTING
    method.
    """

    print("")
    print("===============================")
    print("BACKHAUL ROUTING SOLVER EXAMPLE")
    print("===============================")
    print("")

    # Clear any related tables in case they already exist
    kinetica.clear_table(
        table_name=GRAPH_S_BSOLVED,
        options=OPTION_NO_ERROR
    )

    # Solve GRAPH_S for Backhaul Routing using the given source node and
    # routing to the given destination nodes
    print("Solving {} using BACKHAUL_ROUTING solver type.".format(GRAPH_S))
    fixed_assets = [
        "POINT(-122.37694377950754 47.69058183874783)",
        "POINT(-122.37591381124582 47.6947415132984)",
        "POINT(-122.3635541921052 47.70005617015495)",
        "POINT(-122.35565776876535 47.70583235665686)",
        "POINT(-122.35531444601145 47.72477375603093)",
        "POINT(-122.33540172628489 47.72361898977214)",
        "POINT(-122.32407207540598 47.72408089934756)",
        "POINT(-122.32647533468332 47.73562730759159)",
        "POINT(-122.32922191671457 47.74209217807247)",
        "POINT(-122.31205577901926 47.741399551774876)",
        "POINT(-122.31274242452707 47.737705389219734)",
        "POINT(-122.29935283712473 47.73701270455902)",
        "POINT(-122.2976362233552 47.73262548770682)",
        "POINT(-122.2921430592927 47.73354914302527)",
        "POINT(-122.29179973653879 47.72200227400312)",
        "POINT(-122.30003948263254 47.712531931184785)",
        "POINT(-122.30278606466379 47.701442513285734)",
        "POINT(-122.30518932394114 47.694048257248284)",
        "POINT(-122.30244274190989 47.68827076507089)",
        "POINT(-122.31514568380442 47.68480396253861)",
        "POINT(-122.32750530294504 47.68711518982914)",
        "POINT(-122.3474180226716 47.687577422998146)",
        "POINT(-122.37763042501535 47.686421832395006)",
        "POINT(-122.37694377950754 47.69058183874783)"
    ]
    remote_assets = [
        "POINT(-122.324612 47.725231)", "POINT(-122.362112 47.696131)",
        "POINT(-122.356212 47.74173099999999)", "POINT(-122.341512 47.702631)",
        "POINT(-122.327412 47.737131)", "POINT(-122.344912 47.727531)",
        "POINT(-122.291612 47.65763099999999)", "POINT(-122.324512 47.703831)",
        "POINT(-122.297712 47.731231)", "POINT(-122.317912 47.699431)",
        "POINT(-122.351412 47.713731)", "POINT(-122.295612 47.670531)",
        "POINT(-122.321812 47.672731)", "POINT(-122.295612 47.697531)",
        "POINT(-122.333012 47.689831)", "POINT(-122.298812 47.711731)",
        "POINT(-122.309612 47.684931)", "POINT(-122.359712 47.716431)",
        "POINT(-122.360612 47.659231)", "POINT(-122.299012 47.707431)",
        "POINT(-122.311612 47.66683099999999)", "POINT(-122.315912 47.724631)",
        "POINT(-122.316912 47.709531)", "POINT(-122.313212 47.685731)",
        "POINT(-122.321712 47.721231)", "POINT(-122.339512 47.695031)",
        "POINT(-122.310512 47.692531)", "POINT(-122.326512 47.679731)",
        "POINT(-122.358112 47.705731)", "POINT(-122.352812 47.715931)",
        "POINT(-122.291612 47.67753099999999)", "POINT(-122.295712 47.666731)",
        "POINT(-122.303712 47.681531)", "POINT(-122.362312 47.732231)",
        "POINT(-122.347712 47.740431)", "POINT(-122.343512 47.68523099999999)",
        "POINT(-122.326412 47.730631)", "POINT(-122.357812 47.67003099999999)",
        "POINT(-122.321012 47.694531)", "POINT(-122.353312 47.737131)",
        "POINT(-122.323412 47.67903099999999)", "POINT(-122.351012 47.701731)",
        "POINT(-122.349112 47.735931)", "POINT(-122.361612 47.675031)",
        "POINT(-122.320912 47.70193099999999)", "POINT(-122.345412 47.693331)",
        "POINT(-122.357112 47.669631)", "POINT(-122.352912 47.685131)",
        "POINT(-122.361712 47.681331)", "POINT(-122.346312 47.722231)"
    ]
    solve_s_bgraph_response = kinetica.solve_graph(
        graph_name=GRAPH_S,
        solver_type="BACKHAUL_ROUTING",
        source_nodes=fixed_assets,
        destination_nodes=remote_assets,
        solution_table=GRAPH_S_BSOLVED,
        options={"export_solve_results": "true"}
    )["result_per_destination_node"]
    print("Cost for each remote asset to travel to nearest fixed asset:")
    print("Remote Asset                           Cost (in minutes)")
    print("=====================================  =================")
    for asset, cost in zip(remote_assets, solve_s_bgraph_response):
        print("{:<37s}  {:<11f}".format(asset, (cost / 60)))

# end backhaul_example()

if __name__ == '__main__':
    # Establish connection with a locally-running instance of Kinetica
    kinetica = gpudb.GPUdb(encoding='BINARY', host=HOST, port=PORT)

    # Execute defined functions
    table_setup()
    graph_setup()
    backhaul_example()